/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping;

import static java.lang.Math.max;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.constraint.ConstrainedGenotypeMap;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;

import com.google.inject.Inject;

/**
 * Mutate operator manipulates the allocation of tasks to resources by replacing an existing
 * allocation with a new random allocation.
 * 
 * @author huang
 */
public class MutateAllocation<S extends TaskMappingEntry<ITaskAdapter<?>, IExecutionUnitAdapter<?>>, T extends TaskMappingEncoding<ITaskAdapter<?>, S>>
		extends MutateTasksBase<S, T> {
	/** Constructs a new allocation mutation operator for a given exploration run */
	@Inject
	public MutateAllocation(Collection<ITaskAdapter<?>> tasks, ConstrainedGenotypeMap constraints) {
		super(tasks, constraints);
	}

	/** Constant to enable / disable replication count check. */
	private static final boolean CHECK_REPLICATION_COUNT = false;

	/** {@inheritDoc} */
	@Override
	public void mutate(T encoding, double rate) {
		checkReplicationCount(encoding);

		Collection<ITaskAdapter<?>> encDeplComponents = new ArrayList<>(encoding.getRequesters());
		for(ITaskAdapter<?> deployableComponent : encDeplComponents) {
			if(rand.nextDouble() < rate) {
				Collection<S> assignedResources = encoding.getMappingEntriesOf(deployableComponent);
				if(assignedResources == null) {
					return;
				}
				// Perform the re-allocation
				try {
					reAllocateMappingEntry(encoding, deployableComponent);
				} catch(ExplorationException e) {
					// Opt4J does not allow to throw Exceptions here.
					throw new RuntimeException(e);
				}
				checkReplicationCount(encoding);
			}
		}

		checkReplicationCount(encoding);
	}

	/** Checks the replication count and prints a message in case of errors. */
	private void checkReplicationCount(T encoding) {
		if(CHECK_REPLICATION_COUNT) {
			for(ITaskAdapter<?> currentComponent : tasks) {
				if(encoding.getAllocatedResourcesFor(currentComponent).size() != 1) {
					System.out.println("Encoding replications exceed the allowed amount");
				}
			}
		}
	}

	/**
	 * Performs the actual re-allocation of the given {@link ITaskAdapter}.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	private void reAllocateMappingEntry(T encoding, ITaskAdapter<?> deployableComponent)
			throws ExplorationException {
		// The assigned resources must be an indexed list, such that randomly determined allocations
		// can be directly removed using a random number generator.
		List<S> assignedResources =
				new ArrayList<>(encoding.getMappingEntriesOf(deployableComponent));
		if(assignedResources.isEmpty()) {
			return;
		}

		// Reassign a random allocation...
		// max(...) is required as nextInt does not support a bound of 0.
		int selIdx = max(0, rand.nextInt(assignedResources.size()) - 1);
		S selectedEntry = assignedResources.get(selIdx);

		// This respects {@link ExcludeDeploymentConstraint}s.
		if(encoding.hasLimitedDeployment(deployableComponent)) {
			// List<IExecutionUnitAdapter<?>> physExecUnits =
			// new ArrayList<>(encoding.getAllowedAllocations(deployableComponent));
			// List<Partition> targets = encoding.getAllowedPartitionTargets(physExecUnits);
			// TODO: avoid the raw casting.
			List<Partition> targets = new ArrayList<>((Collection<Partition>)(Collection<?>)encoding
					.getAllowedAllocations(deployableComponent));
			int selExecU = max(0, rand.nextInt(targets.size()) - 1);
			selectedEntry.resetTarget(targets.get(selExecU));
		} else {
			List<Partition> targets = (List<Partition>)(Collection<?>)encoding
					.getAllowedAllocations(deployableComponent);
			int selExecU = max(0, rand.nextInt(targets.size()) - 1);
			selectedEntry.resetTarget(targets.get(selExecU));
		}
	}
}
