/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.convert;

import static org.fortiss.tooling.common.util.LambdaUtils.filter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * Base class for implementing simple converters for (parts of) AF3 elements into DDD-SE-internal
 * representations. Provides management code for the single {@link IConverterModule}s. All
 * {@code null}s in the input set are silently ignored.
 * 
 * @author diewald
 */
public abstract class ConversionManagerBase<I, O> {

	/** Caches the input set. */
	private Collection<I> inputSet;

	/** Output set of transformed elements. */
	private Map<I, O> outputSet = new HashMap<>();

	/** Set of {@link IConverterModule}s registered in this conversion manager. */
	protected Collection<IConverterModule<I, O>> converters = new ArrayList<>();

	/** Constructor. */
	public ConversionManagerBase(Collection<I> inputSet) {
		this.inputSet = filter(inputSet, e -> e != null);
		addConverters();
	}

	/** Registers the listed set of {@link IConverterModule}s. */
	protected abstract void addConverters();

	/**
	 * Performs the conversion of the input to the output set if applicable.
	 * 
	 * @throws Exception
	 *             if the conversion fails (badly) such that DSE cannot proceed successfully
	 *             anymore.
	 */
	public Map<I, O> convert() throws Exception {
		for(I inputElement : inputSet) {
			for(IConverterModule<I, O> converter : converters) {
				O convResult = converter.convert(inputElement);
				if(null != convResult) {
					outputSet.put(inputElement, convResult);
					// Found a matching converter, no need to try the others.
					continue;
				}
			}
		}
		return outputSet;
	}
}
