/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.MappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.dsl_v2.model.booleanp.Implies;
import org.fortiss.af3.exploration.dsl_v2.model.booleanp.allocation.Allocation;
import org.fortiss.af3.exploration.dsl_v2.model.booleanp.allocation.Dislocation;
import org.fortiss.af3.exploration.dsl_v2.model.expression.IExpression;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * An expression evaluator which is able to handle the location operators and the internal mapping
 * representation.
 * 
 * @author diewald
 */
public abstract class TaskMappingParser<R> extends ExtendedExplorationExpressionEvaluatorBase {

	/** Reference to the mapping which shall be evaluated here. */
	protected MappingEncoding<?, ITaskAdapter<?>, IExecutionUnitAdapter<?>, ?> mapping;

	/** Mapping of Partitions to physical execution units. */
	// FIXME: HACK in order to validate the mapping of components/tasks to phys exec units if
	// partitions are used.
	protected MappingEncoding<?, IRequestAdapter<?>, IExecutionUnitAdapter<?>, ?> virtMapping;

	/** The {@link IExpression} to be parsed. */
	protected IExpression expression;

	/**
	 * {@link TaskMapper} for association AF3 task references with the internal {@link ITaskAdapter}
	 * elements.
	 */
	protected TaskMapper<EObject> taskMapper;

	/** Constructor. */
	@SuppressWarnings("unchecked")
	public TaskMappingParser(
			MappingEncoding<?, ITaskAdapter<?>, IExecutionUnitAdapter<?>, ?> mapping,
			MappingEncoding<?, ? extends IRequestAdapter<?>, IExecutionUnitAdapter<?>, ?> virtMapping,
			IExpression expression, TaskMapper<EObject> taskMapper) {
		this.mapping = mapping;
		this.virtMapping =
				(MappingEncoding<?, IRequestAdapter<?>, IExecutionUnitAdapter<?>, ?>)virtMapping;
		this.expression = expression;
		this.taskMapper = taskMapper;
	}

	/** Entry point for parsing {@link IExpression}s containing allocation {@link IExpression}s. */
	public abstract R parse() throws Exception;

	/** {@inheritDoc} */
	@Override
	public Object visit(Allocation alloc) throws Exception {
		// TODO: Here, we could use the requester / resource framework for the return types. In
		// addition we would obtain a generic mapping evaluator.
		IModelElement requester = alloc.getLeft().accept(this, IModelElement.class);
		IModelElement resource = alloc.getRight().accept(this, IModelElement.class);

		IRequestAdapter<?> requestAdp = taskMapper.getAdapter(requester);

		Collection<IExecutionUnitAdapter<?>> allocResourceAdps =
				mapping.getAllocatedResourcesFor((ITaskAdapter<?>)requestAdp);
		for(IResourceAdapter<?> allocResource : allocResourceAdps) {
			if(allocResource.getObject() == resource) {
				return Boolean.valueOf(true);
			} else if(allocResource instanceof IRequestAdapter) {
				// Fallback if a virtualization layer exists.
				// "get(0)": Partitions are not replicated: There is a n:1 mapping of Parititons to
				// execution units.
				IExecutionUnitAdapter<?> physResource = getFirst(
						virtMapping.getAllocatedResourcesFor((IRequestAdapter<?>)allocResource))
								.get();
				if(physResource.getObject() == resource) {
					return Boolean.valueOf(true);
				}
			}
		}

		return Boolean.valueOf(false);
	}

	/** {@inheritDoc} */
	@Override
	public Object visit(Dislocation disLoc) throws Exception {
		IModelElement requester = disLoc.getLeft().accept(this, IModelElement.class);
		IModelElement resource = disLoc.getRight().accept(this, IModelElement.class);

		IRequestAdapter<?> requestAdp = taskMapper.getAdapter(requester);

		Collection<IExecutionUnitAdapter<?>> allocResourceAdps =
				mapping.getAllocatedResourcesFor((ITaskAdapter<?>)requestAdp);
		for(IResourceAdapter<?> allocResource : allocResourceAdps) {
			if(allocResource.getObject() == resource) {
				return Boolean.valueOf(false);
			}
		}

		return Boolean.valueOf(true);
	}

	/** {@inheritDoc} */
	@Override
	public Object visit(Implies implies) throws Exception {
		// FIXME: How to implement implies?
		return null;
	}
}
