/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.EMFModelConcurrencyHandler;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationTransformationInputs;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.moea.model.annotation.AnnotationFactory;
import org.fortiss.af3.exploration.moea.model.annotation.ImplementsComponentRef;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;

/**
 * 
 * @author diewald
 */
public class ComponentArchitectureFinalizer
		extends TransformationModuleBase<ComponentArchitecture> {

	/** {@inheritDoc} */
	@Override
	public Class<ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		inputTypes.add(InstantiatedTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ComponentArchitectureTransformer.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.FINALIZING;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws TransformationModuleException
	 *             see {@link TransformationModuleBase}.
	 */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			TransformationState state, TransformationContext context,
			ExplorationTransformationInputs baseModels, SubClassToInstanceMap<N> transformedModels,
			EObject outputModel) throws TransformationModuleException {
		// Install the back reference specification to the components in the "150%/125% model".
		try {
			createImplementsComponentSpecifications(
					((ComponentArchitecture)outputModel).getTopComponent());
		} catch(InvalidTimingModelException e) {
			throw new TransformationModuleException(this,
					"Could not identify the relevant components, probably due to an inconsistent timing specification.");
		}

		// instantiateAnnotationsRecursive((ComponentArchitecture)outputModel);

		// Re-enable notifications for the component architecture representing the replica.
		EMFModelConcurrencyHandler.getInstance().tryEnableNotificationsSyncOf(outputModel);

		return (ComponentArchitecture)outputModel;
	}

	/**
	 * Creates a {@link ImplementsComponentRef} {@link IModelElementSpecification} foreach
	 * {@link Component}-value given in the map that references the corresponding {@link Component}
	 * -key. Thereby it is possible to identify the implemented {@link Component} (key) from the
	 * instantiated {@link Component} (value).
	 * 
	 * @throws InvalidTimingModelException
	 */
	private void createImplementsComponentSpecifications(Component component)
			throws InvalidTimingModelException {
		for(Component curComp : component.getSubComponents()) {
			EObject originalComponent = genToOrigAssocMap.get(curComp);

			InstantiatedTaskMappingEntry entry = null;
			if(originalComponent instanceof Component) {
				entry = (InstantiatedTaskMappingEntry)encEntryModelElemAssoc.inverse()
						.get(originalComponent, InstantiatedTaskMappingEntry.class);
			}

			if(entry != null && entry.getSource().isInstantiation()) {
				addImplementsComponentRefSpec(curComp, (Component)originalComponent);
			} else {
				createImplementsComponentSpecifications(curComp);
			}
		}
	}

	/**
	 * Creates a new {@link ImplementsComponentRef} {@link IModelElementSpecification} for the given
	 * {@code element} which references the given {@code referencedComponent}.
	 */
	private void addImplementsComponentRefSpec(Component element, Component referencedComponent) {
		List<IModelElementSpecification> specList = element.getSpecifications();

		ImplementsComponentRef refSpec = AnnotationFactory.eINSTANCE.createImplementsComponentRef();
		refSpec.setRefComponent(referencedComponent);

		specList.add(refSpec);
	}
}
