/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static com.google.common.base.Preconditions.checkArgument;
import static org.eclipse.emf.ecore.util.EcoreUtil.getAllContents;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;

import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.conqat.lib.commons.reflect.ReflectionUtils;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.IProblemDimension;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.IVirtualizationPlatformArchitectureElement;
import org.fortiss.tooling.base.model.element.IHierarchicElement;

import com.google.common.collect.Multimap;

/**
 * Static utility methods mainly used in the Design Space Exploration.
 * 
 * @author diewald
 */
public class ExplorationAlgUtils {

	/** Constructor. Prevent initialization. */
	private ExplorationAlgUtils() {
	}

	/**
	 * Returns the given Collection as a Collection of the raw types of its elements.
	 * 
	 * @param origCollection
	 *            Collection to be casted.
	 * @return Collection of the raw types of the input collection, or {@code null} if the
	 *         Collection cannot be created.
	 */
	// FIXME: Remove this method, it is unsafe and cannot be used with instances of EList which have
	// parameterized constructors (Factory pattern).
	@SuppressWarnings("unchecked")
	public static <C extends Collection<T>, T, CR extends Collection<? super T>> CR
			castToRawColl(C origCollection) {
		CR castedCollection = (CR)Collections.emptySet();
		if(origCollection != null && !origCollection.isEmpty()) {
			try {
				Constructor<CR> ctor =
						(Constructor<CR>)origCollection.getClass().getDeclaredConstructor();
				castedCollection = ctor.newInstance();
				((Collection<T>)castedCollection).addAll(origCollection);
			} catch(InstantiationException | IllegalAccessException | IllegalArgumentException |
					InvocationTargetException | NoSuchMethodException | SecurityException e) {
				// Never here.
				e.printStackTrace();
			}
		}
		return castedCollection;
	}

	/**
	 * Returns the given Collection as a Collection of the raw types of its elements.
	 * 
	 * @param origCollection
	 *            Collection to be casted.
	 * @param castedCollType
	 *            Collection type that shall be used to collect the casted values.
	 * @return Collection of the raw types of the input collection, or {@code null} if the
	 *         Collection cannot be created.
	 */
	// FIXME: Remove this method, it is unsafe and cannot be used with instances of EList which have
	// parameterized constructors (Factory pattern).
	@SuppressWarnings("unchecked")
	public static <C extends Collection<T>, T, CR extends Collection<? super T>> CR
			castToRawColl(C origCollection, Class<CR> castedCollType) {
		CR castedCollection = (CR)Collections.emptySet();
		if(origCollection != null && !origCollection.isEmpty()) {
			try {
				Constructor<CR> ctor = castedCollType.getDeclaredConstructor();
				castedCollection = ctor.newInstance();
				((Collection<T>)castedCollection).addAll(origCollection);
			} catch(InstantiationException | IllegalAccessException | NoSuchMethodException |
					SecurityException | IllegalArgumentException | InvocationTargetException e) {
				// Never here.
				e.printStackTrace();
			}
		}
		return castedCollection;
	}

	/**
	 * For a given Object, determine whether it is a sub-class of any of the classes contained in
	 * the given collection.
	 */
	public static boolean isAssignableFromAny(Object obj, Collection<Class<?>> classes) {
		return classes.stream().anyMatch(cls -> cls.isAssignableFrom(obj.getClass()));
	}

	/**
	 * Finds the (target) {@link InputPort}s to which the given {@link OutputPort} is connected.
	 * 'Target Ports' are connected {@link InputPort}s of {@link Component}s from the set of
	 * {@code deployableComponents}.
	 * 
	 * @param sourcePort
	 *            The port to start the search from.
	 * @param deployableComponents
	 *            Collection of 'deployable' {@link Component}s present in the input logical
	 *            architecture.
	 * @return The list of target ports of atomic {@link Component}s.
	 */
	public static List<Port> findTargetPorts(OutputPort sourcePort,
			Collection<Component> deployableComponents) {
		if(!deployableComponents.contains(sourcePort.getComponent())) {
			return Collections.emptyList();
		}
		List<Port> targets = new ArrayList<>();
		LinkedList<Channel> exploreChannels = new LinkedList<Channel>();
		exploreChannels.addAll(sourcePort.getOutgoingChannels());
		while(!exploreChannels.isEmpty()) {
			Channel currentExploreChannel = exploreChannels.removeFirst();
			Component currentTargetComponent = currentExploreChannel.getTarget().getComponent();
			if(deployableComponents.contains(currentTargetComponent)) {
				targets.add(currentExploreChannel.getTarget());
			} else {
				exploreChannels.addAll(currentExploreChannel.getTarget().getOutgoingChannels());
			}
		}
		return targets;
	}

	/**
	 * Returns a list of source ports within the given {@link ComponentArchitecture}. Source ports
	 * are defined as ports that do not have direct inputs.
	 * 
	 * @param componentArchitecture
	 *            The {@link ComponentArchitecture} for whom source ports shall be identified.
	 * @return {@link EList} of source {@link OutputPort}s.
	 * @throws InvalidTimingModelException
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	public static EList<OutputPort> getSourcePorts(ComponentArchitecture componentArchitecture)
			throws InvalidTimingModelException {
		if(componentArchitecture == null) {
			return null;
		}

		EList<OutputPort> allOutputPortsInArch =
				getChildrenWithType(componentArchitecture, OutputPort.class);
		EList<OutputPort> sourcePortList = new BasicEList<OutputPort>();
		for(OutputPort currentOutputPort : allOutputPortsInArch) {
			// Filter out ports that represent the transition through a hierarchic hull.
			if(currentOutputPort.getIncoming().isEmpty() /*
															 * &&
															 * !isAnyParentComponentDeployable(
															 * currentOutputPort.getComponent())
															 */) {
				// FIXME: HERE
				sourcePortList.add(currentOutputPort);
			}
		}

		return sourcePortList;
	}

	/**
	 * Returns the set of {@link OutputPort}s of all given 'deployable' {@link Component}s.
	 * 
	 * @param deployableComponents
	 *            Collection of all {@link OutputPort}s of 'deployable' {@link Component}s.
	 * @return List of 'deployable' {@link OutputPort}s or an empty list if none can be identified.
	 */
	public static EList<OutputPort>
			getDeployableOutputPorts(Collection<Component> deployableComponents) {
		EList<OutputPort> sourcePortList = new BasicEList<OutputPort>();

		for(Component component : deployableComponents) {
			sourcePortList.addAll(component.getOutputPorts());
		}

		return sourcePortList;
	}

	/**
	 * Returns all Objects from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceColl
	 *            Collection of Objects to be filtered.
	 * @return Collection of objects inheriting from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<T> pickInstanceOf(Class<T> targetClass, Collection<?> sourceColl) {
		Collection<T> filteredColl = new ArrayList<>();
		for(Object srcObj : sourceColl) {
			if(targetClass.isAssignableFrom(srcObj.getClass())) {
				filteredColl.add((T)srcObj);
			}
		}
		return filteredColl;
	}

	/**
	 * Returns the first found object from the given collection that inherits from the given class.
	 * 
	 * @param targetClass
	 *            Type to check the inheritance of the collection's objects.
	 * @param sourceColl
	 *            Objects to examine.
	 * @return The first found object inheriting from the given class, or {@code null} if none is
	 *         found.
	 */
	@SuppressWarnings("unchecked")
	public static <T> T pickFirstInstanceOf(Class<T> targetClass, Collection<?> sourceColl) {
		for(Object srcObj : sourceColl) {
			if(targetClass.isAssignableFrom(srcObj.getClass())) {
				return (T)srcObj;
			}
		}
		return null;
	}

	/**
	 * Returns all Classes from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceTypeColl
	 *            Collection of Classes to be filtered.
	 * @return Collection of Classes inheriting from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<Class<T>> pickSubTypesOf(Class<T> targetClass,
			Collection<Class<?>> sourceTypeColl) {
		Collection<Class<T>> filteredColl = new ArrayList<>();
		for(Class<?> srcType : sourceTypeColl) {
			if(targetClass.isAssignableFrom(srcType)) {
				filteredColl.add((Class<T>)srcType);
			}
		}
		return filteredColl;
	}

	/**
	 * Returns all Objects from the given Collection which inherit from the given collection of
	 * classes.
	 * 
	 * @param filterClasses
	 *            Class providing the type filer.
	 * @param unfilteredCollection
	 *            Collection of Objects to be filtered.
	 * @return Collection of Classes that do not inherit from the given class.
	 */
	public static <T, S extends T> Collection<T> filterByTypeUnsafe(
			Collection<Class<?>> filterClasses, Collection<T> unfilteredCollection) {
		if(unfilteredCollection == null) {
			return null;
		}

		Collection<T> filteredCollection = new ArrayList<>();
		for(T element : unfilteredCollection) {
			if(filterClasses.stream().anyMatch(c -> c.isAssignableFrom(element.getClass()))) {
				filteredCollection.add(element);
			}
		}
		return filteredCollection;
	}

	/**
	 * Filters all Classes from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceTypeColl
	 *            Collection of Classes to be filtered.
	 * @return Collection of Classes that do not inherit from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<Class<T>> filterOutSubTypesOf(Class<T> targetClass,
			Collection<Class<?>> sourceTypeColl) {
		Collection<Class<T>> filteredColl = new ArrayList<>();
		for(Class<?> srcType : sourceTypeColl) {
			if(!targetClass.isAssignableFrom(srcType)) {
				filteredColl.add((Class<T>)srcType);
			}
		}
		return filteredColl;
	}

	/**
	 * Determines whether an {@link IPlatformResource} is a virtual platform element (like a
	 * partition).
	 * 
	 * @param element
	 *            The element to be examined.
	 * @return see above
	 */
	public static boolean isVirtualResource(IHierarchicElement element) {
		if(element instanceof IVirtualizationPlatformArchitectureElement) {
			return true;
		}

		return false;
	}

	/**
	 * Returns a list of child elements of the given type for each parent in the given collection.
	 * NOTE: This operation can be very costly, if elements are contained in the list which are far
	 * up in the hierarchy.
	 * 
	 * @param parents
	 *            Collection of parents to be examined.
	 * @param type
	 *            The class to be filter for.
	 * @return List of child elements of the given type.
	 */
	public static <S> EList<S> getAllChildrenWithType(Collection<? extends EObject> parents,
			Class<S> type) {
		EList<S> childrenWithType = new BasicEList<S>();
		for(EObject currentParent : parents) {
			childrenWithType.addAll(getChildrenWithType(currentParent, type));
		}
		return childrenWithType;
	}

	/**
	 * Determines whether the given hardware element is virtualized by a
	 * {@link IVirtualizationPlatformArchitectureElement}.
	 * 
	 * @param element
	 *            The hardware platform element to be examined.
	 * @param virtualizationProviders
	 *            The (complete) collection of {@link IVirtualizationPlatformArchitectureElement}s
	 *            present in the system.
	 * @return Boolean indicating the presence of a
	 *         {@link IVirtualizationPlatformArchitectureElement} connected to the given
	 *         hardware element.
	 */
	public static boolean hasVirtualizationLayer(IPlatformResource element,
			Collection<IVirtualizationPlatformArchitectureElement> virtualizationProviders) {
		Collection<IPlatformResource> selfAndParentElements = new HashSet<>();
		selfAndParentElements.add(element);
		selfAndParentElements.addAll(getParentsWithType(element, IPlatformResource.class));

		Collection<IPlatformResource> virtualizedResources = new HashSet<IPlatformResource>();
		for(IVirtualizationPlatformArchitectureElement currentVirtualizationProvider : virtualizationProviders) {
			// TODO: Replace by Allocation Model
			// virtualizedResources.addAll(PlatformArchitectureUtils
			// .getResourceReferences((IModelElement)currentVirtualizationProvider));
		}

		return (Collections.disjoint(selfAndParentElements, virtualizedResources)) ? false : true;
	}

	/**
	 * Determines whether the given {@link IHierarchicElement}s are contained within the
	 * same (parent) {@link IHierarchicElement}.
	 * 
	 * @param elementA
	 *            The first {@link IHierarchicElement} to be checked.
	 * @param elementB
	 *            The {@link IHierarchicElement} to be checked against.
	 * @return see above.
	 */
	public static boolean haveSameParentElement(IHierarchicElement elementA,
			IHierarchicElement elementB) {
		return (elementA.getContainer() == elementB.getContainer()) ? true : false;
	}

	/**
	 * Determines whether the given {@link IHierarchicElement}s are contained within the
	 * same (parent) {@link IHierarchicElement}s. This method is an extension to {#link
	 * haveSameParentElement(IHierarchicElement, IHierarchicElement)} which also considers the case
	 * where the given elements can be part of a virtual platform.
	 * 
	 * @param elementA
	 *            The first {@link IHierarchicElement} to be checked.
	 * @param elementB
	 *            The {@link IHierarchicElement} to be checked against.
	 * @param virtualizationProviders
	 *            The (complete) collection of {@link IVirtualizationPlatformArchitectureElement}s
	 *            present in the
	 *            system.
	 * @return see above.
	 */
	public static boolean haveSameParentHardwareElement(IHierarchicElement elementA,
			IHierarchicElement elementB,
			Collection<IVirtualizationPlatformArchitectureElement> virtualizationProviders) {
		List<IHierarchicElement> refA = null;
		// TODO: Replace by Allocation Model
		// getReferencedElementsWithType(elementA, IHierarchicElement.class);
		List<IHierarchicElement> refB = null;
		// TODO: Replace by Allocation Model
		// getReferencedElementsWithType(elementB, IHierarchicElement.class);
		if(refA.size() < 1 || refB.size() < 1) {
			return false;
		}
		// get(0): Checked above.
		elementA = refA.get(0);
		// get(0): Checked above.
		elementB = refB.get(0);
		return (haveSameParentElement(elementA, elementB)) ? true : false;
	}

	/**
	 * Creates a filtered list from a given list of superclasses of the filter class.
	 * 
	 * @param generalList
	 *            The list of superclasses of {#link filterClass} which will be filtered.
	 * @param filterClass
	 *            The class by which the given list shall be filtered
	 * @return The filter list with elements of the type T
	 */
	@SuppressWarnings("unchecked")
	public static <S, T extends S> EList<T> getElementsWithType(EList<S> generalList,
			Class<T> filterClass) {
		EList<T> filteredList = new BasicEList<T>();
		for(S listElement : generalList) {
			if(filterClass.isAssignableFrom(listElement.getClass())) {
				filteredList.add((T)listElement);
			}
		}

		return filteredList;
	}

	/**
	 * Returns the List of <i>all</i> {@link TimeSlot}s that are predecessors of the given
	 * {@link TimeSlot}. Therefore, this method evaluates the predecessor field of {@link TimeSlot}
	 * s.
	 * 
	 * @param timeSlot
	 *            The {@link TimeSlot} for which the predecessors shall be identified.
	 * @return List of all predecessor {@link TimeSlot}s.
	 */
	public static List<TimeSlot> getAllPredecessorsOfTimeSlot(TimeSlot timeSlot) {
		List<TimeSlot> predescessorList = new ArrayList<TimeSlot>();
		Queue<TimeSlot> toBeVisited = new LinkedList<TimeSlot>();
		toBeVisited.addAll(timeSlot.getPredecessors());
		while(!toBeVisited.isEmpty()) {
			TimeSlot currentPredecessor = toBeVisited.poll();
			if(!predescessorList.contains(currentPredecessor)) {
				predescessorList.add(currentPredecessor);
				toBeVisited.addAll(currentPredecessor.getPredecessors());
			}
		}

		return predescessorList;
	}

	/**
	 * Searches for a method annotated with the given {@link Annotation} type in the given class and
	 * returns it. If no such method is found, {@code null} is returned; if more than one annotated
	 * method is found, an exception is thrown.
	 * 
	 * @param containingClass
	 *            Class to be examined.
	 * @param annotationType
	 *            annotation of the seeked method.
	 * @return see above.
	 */
	public static Method getMethodAnnotatedWith(Class<?> containingClass,
			Class<? extends Annotation> annotationType) {
		// From Google Guice --> Matchers.
		Retention retention = annotationType.getAnnotation(Retention.class);
		checkArgument(retention != null && retention.value() == RetentionPolicy.RUNTIME,
				"Annotation %s is missing RUNTIME retention", annotationType.getSimpleName());
		// END Matchers snippet.
		Method annotatedMethod = null;
		for(Method method : containingClass.getMethods()) {
			if(method.getAnnotationsByType(annotationType).length != 0) {
				if(annotatedMethod == null) {
					// First match.
					annotatedMethod = method;
				} else {
					// Multiple methods with the given annotation found.
					throw new ExplorationRuntimeException("Found multiple methods in the class " +
							containingClass.getSimpleName() + " annotated with " +
							annotationType.getSimpleName() + ".");
				}
			}
		}
		return annotatedMethod;
	}

	/**
	 * Returns the List of <i>all</i> {@link TimeSlot}s that are successors of the given
	 * {@link TimeSlot}. Therefore, this method evaluates the successor field of {@link TimeSlot}s.
	 * 
	 * @param timeSlot
	 *            The {@link TimeSlot} for which the successors shall be identified.
	 * @return List of all successor {@link TimeSlot}s.
	 */
	public static List<TimeSlot> getAllSuccessorsOfTimeSlot(TimeSlot timeSlot) {
		List<TimeSlot> successorList = new ArrayList<TimeSlot>();
		Queue<TimeSlot> toBeVisited = new LinkedList<TimeSlot>();
		toBeVisited.addAll(timeSlot.getSuccessors());
		while(!toBeVisited.isEmpty()) {
			TimeSlot currentPredecessor = toBeVisited.poll();
			if(!successorList.contains(currentPredecessor)) {
				successorList.add(currentPredecessor);
				toBeVisited.addAll(currentPredecessor.getSuccessors());
			}
		}

		return successorList;
	}

	/**
	 * Returns a collection of all child {@link EObject}s that have either of the given types.
	 * 
	 * @param parent
	 *            {@link EObject} to start the search from.
	 * @param typeList
	 *            filter types for the child elements.
	 * @param isReflexive
	 *            if the parent shall be checked as well.
	 * @return Collection of child {@link EObject}s conforming to either of the given types.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> Collection<T> getChildrenWithTypes(EObject parent,
			Collection<Class<?>> typeList, boolean isReflexive) {
		Collection<T> children = new HashSet<T>();
		if(isReflexive && isAssignableFromAny(parent, typeList)) {
			children.add((T)parent);
		}

		TreeIterator<EObject> content = getAllContents(parent, true);
		while(content.hasNext()) {
			EObject child = content.next();
			if(isAssignableFromAny(child, typeList)) {
				children.add((T)child);
			}
		}
		return children;
	}

	/**
	 * Debug method to search collections for strings.
	 * 
	 * @param searchString
	 *            The string for which the elements in the collection shall be evaluated against.
	 * @param objectCollection
	 *            The collection of objects to be examined
	 * @return boolean, if any object's toString method contains the search string.
	 */
	public static boolean containsElementByString(String searchString,
			Collection<?> objectCollection) {
		for(Object obj : objectCollection) {
			if(obj.toString().contains(searchString)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns the superclass of the given element which implements the given interface.
	 * 
	 * @param element
	 *            The element whose implementing interface class shall be returned.
	 * @param superClass
	 *            The interface class for which the superclasses of {#link element} shall be
	 *            examined
	 * @return The first identified instance of a superclass implementing {#link superClass}, or
	 *         {#code null} if none was found.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Class<T> getSuperClassOfImplementingInterface(EObject element,
			Class<T> superClass) {
		EList<EClass> superClasses = element.eClass().getESuperTypes();
		for(EClass currentSuperClass : superClasses) {
			if(superClass.isAssignableFrom(currentSuperClass.getInstanceClass())) {
				return (Class<T>)currentSuperClass.getInstanceClass();
			}
		}
		return null;
	}

	/**
	 * Determines the {@link IProblemDimension} of the given {@link ExplorationTarget}.
	 * 
	 * @param goal
	 *            The {@link ExplorationTarget} whose {@link IProblemDimension} shall be identified.
	 * @return The {@link IProblemDimension} of the given {@link ExplorationTarget} or {#code null}
	 *         if none is identified.
	 */
	public static <T extends IProblemDimension> Class<IProblemDimension>
			getAssociatedDimensionFromGoal(ExplorationTarget<?> goal) {
		return getSuperClassOfImplementingInterface(goal, IProblemDimension.class);
	}

	/**
	 * Returns the typed key from the map which is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is {@link Object}, which is
	 * considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire inheritance tree) the
	 * results should be cached if multiple lookups for the same class are expected.
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param keySet
	 *            the set of keys to perform the lookup in.
	 * @return the best match found or <code>null</code> if no matching key
	 *         was found. Note that <code>null</code> will also be returned if
	 *         the entry for the best matching class was <code>null</code>.
	 * 
	 *         Adapted from {@link ReflectionUtils#performNearestClassLookup(Class, Map)}.
	 */
	@SuppressWarnings("unchecked")
	public static <K> Class<? extends K> performNearestClassLookup(Class<? extends K> clazz,
			Set<Class<? extends K>> keySet) {
		Queue<Class<? extends K>> q = new LinkedList<Class<? extends K>>();
		q.add(clazz);

		Class<? extends K> current = null;
		while(!q.isEmpty()) {
			current = q.poll();
			if(keySet.contains(current)) {
				return current;
			}

			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add((Class<? extends K>)superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add((Class<? extends K>)iface);
			}
		}
		return current;
	}

	/**
	 * Returns the value from the {@link Multimap}, whose key is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is {@link Object}, which is
	 * considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire inheritance tree) the
	 * results should be cached if multiple lookups for the same class are expected.
	 * 
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param classMap
	 *            the {@link Multimap} to perform the lookup in.
	 * @return the best match found or the empty set if no matching entry
	 *         was found. Note that the empty set will also be returned if
	 *         the entry for the best matching class was the empty set.
	 * 
	 *         Adapted from {@link ReflectionUtils#performNearestClassLookup(Class, Map)}.
	 */
	public static <K, T> Collection<T> performNearestClassLookup(Class<? extends K> clazz,
			Multimap<Class<? extends K>, T> classMap) {
		Class<? extends K> key = performNearestClassLookup(clazz, classMap.keySet());
		return classMap.get(key);
	}

	/**
	 * Takes a checked Exception and allows its usage in a RuntimeException based environment.
	 * 
	 * @param exception
	 *            Exception to wrap.
	 * @return not used, only rethrows the Exception.
	 * @throws T
	 *             the original Exception.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Exception> RuntimeException rethrow(Exception exception) throws T {
		throw(T)exception; // rely on vacuous cast
	}
}
